""" Module to test the cf_logging library """
import uuid
import logging
import time
import threading
from json import JSONDecoder
import pytest
from json_validator.validator import JsonValidator
from sap import cf_logging
from tests.log_schemas import JOB_LOG_SCHEMA
from tests.util import config_logger

# pylint: disable=protected-access


@pytest.fixture(autouse=True)
def before_each():
    """ reset logging framework """
    cf_logging._SETUP_DONE = False


@pytest.mark.parametrize('log_callback', [
    lambda logger, msg: logger.debug('message: %s', msg),
    lambda logger, msg: logger.info('message: %s', msg),
    lambda logger, msg: logger.warning('message: %s', msg),
    lambda logger, msg: logger.error('message: %s', msg),
    lambda logger, msg: logger.critical('message: %s', msg)
])
def test_log_in_expected_format(log_callback):
    """ Test the cf_logger as a standalone """
    cf_logging.init(level=logging.DEBUG)
    logger, stream = config_logger('cli.test')
    log_callback(logger, 'hi')
    log_json = JSONDecoder().decode(stream.getvalue())
    _, error = JsonValidator(JOB_LOG_SCHEMA).validate(log_json)

    assert error == {}


def test_set_correlation_id():
    """ Test setting correlation_id """
    correlation_id = '1234'
    cf_logging.init(level=logging.DEBUG)
    cf_logging.FRAMEWORK.context.set_correlation_id(correlation_id)

    logger, stream = config_logger('cli.test')
    logger.info('hi')

    log_json = JSONDecoder().decode(stream.getvalue())
    _, error = JsonValidator(JOB_LOG_SCHEMA).validate(log_json)

    assert error == {}
    assert log_json['correlation_id'] == correlation_id
    assert cf_logging.FRAMEWORK.context.get_correlation_id() == correlation_id


def test_exception_stacktrace():
    """ Test exception stacktrace is logged """
    cf_logging.init(level=logging.DEBUG)
    logger, stream = config_logger('cli.test')

    try:
        return 1 / 0
    except ZeroDivisionError:
        logger.exception('zero division error')
        log_json = JSONDecoder().decode(stream.getvalue())
        _, error = JsonValidator(JOB_LOG_SCHEMA).validate(log_json)

        assert error == {}
        assert 'ZeroDivisionError' in str(log_json['stacktrace'])
        assert log_json["msg"] == 'zero division error'


def test_exception_stacktrace_info_level():
    """ Test exception stacktrace is logged """
    cf_logging.init(level=logging.DEBUG)
    logger, stream = config_logger('cli.test')

    try:
        return 1 / 0
    except ZeroDivisionError as exc:
        logger.info('zero division error', exc_info=exc)
        log_json = JSONDecoder().decode(stream.getvalue())
        _, error = JsonValidator(JOB_LOG_SCHEMA).validate(log_json)

        assert error == {}
        assert 'ZeroDivisionError' in str(log_json['stacktrace'])
        assert log_json["msg"] == 'zero division error'


def test_custom_fields_set():
    """ Test custom fields are set up """
    cf_logging.init(level=logging.DEBUG, custom_fields={'cf1': None})
    assert 'cf1' in cf_logging.FRAMEWORK.custom_fields.keys()

def test_thread_safety():
    """ test context keeps separate correlation ID per thread """
    class _SampleThread(threading.Thread):
        def __init__(self):
            super(_SampleThread, self).__init__()
            self.correlation_id = str(uuid.uuid1())
            self.read_correlation_id = ''

        def run(self):
            cf_logging.FRAMEWORK.context.set_correlation_id(self.correlation_id)
            time.sleep(0.1)
            self.read_correlation_id = cf_logging.FRAMEWORK.context.get_correlation_id()

    cf_logging.init(level=logging.DEBUG)

    thread_one = _SampleThread()
    thread_two = _SampleThread()

    thread_one.start()
    thread_two.start()

    thread_one.join()
    thread_two.join()

    assert thread_one.correlation_id == thread_one.read_correlation_id
    assert thread_two.correlation_id == thread_two.read_correlation_id



""" Module that tests the integration of cf_logging with Flask """
import logging
import pytest
from flask import Flask
from flask import Response
from sap import cf_logging
from sap.cf_logging import flask_logging
from tests.log_schemas import WEB_LOG_SCHEMA, JOB_LOG_SCHEMA, CUST_FIELD_SCHEMA
from tests.common_test_params import v_str, v_num, auth_basic, get_web_record_header_fixtures
from tests.util import (
    check_log_record,
    enable_sensitive_fields_logging,
    config_logger,
)


# pylint: disable=protected-access

@pytest.mark.xfail(raises=TypeError, strict=True)
def test_flask_requires_valid_app():
    """ Test the init api expects a valid app """
    flask_logging.init({})


FIXTURE = get_web_record_header_fixtures()
FIXTURE.append(({'Authorization': auth_basic('user', 'pass')},
                {'remote_user': v_str('user')}))
FIXTURE.append(({}, {'response_size_b': v_num(val=2)}))


@pytest.fixture(autouse=True)
def before_each():
    # pylint: disable=duplicate-code
    """ enable all fields to be logged """
    enable_sensitive_fields_logging()
    yield


@pytest.mark.parametrize("headers, expected", FIXTURE)
def test_flask_request_log(headers, expected):
    """ That the expected records are logged by the logging library """

    app = Flask(__name__)

    @app.route('/test/path')
    def _root():
        return Response('ok', mimetype='text/plain')

    _set_up_flask_logging(app)
    _, stream = config_logger('cf.flask.logger')

    client = app.test_client()
    _check_expected_response(client.get('/test/path', headers=headers))
    assert check_log_record(stream, WEB_LOG_SCHEMA, expected) == {}


def test_web_log():
    """ That the custom properties are logged """
    _user_logging({}, {'myprop': 'myval'}, {'myprop': v_str('myval')})


def test_correlation_id():
    """ Test the correlation id is logged when coming from the headers """
    _user_logging({'X-CorrelationID': '298ebf9d-be1d-11e7-88ff-2c44fd152860'},
                  {},
                  {'correlation_id': v_str('298ebf9d-be1d-11e7-88ff-2c44fd152860')})


def test_logging_without_request():
    """ Test logger is usable in non request context """
    app = Flask(__name__)
    _set_up_flask_logging(app)
    logger, stream = config_logger('main.logger')
    logger.info('works')
    assert check_log_record(stream, JOB_LOG_SCHEMA, {'msg': v_str('works')}) == {}

def test_custom_field_loggin():
    """ Test custom fields are generated """
    app = Flask(__name__)
    _set_up_flask_logging(app)
    logger, stream = config_logger('main.logger')
    logger.info('works', extra={'cf1': 'yes'})
    assert check_log_record(stream, CUST_FIELD_SCHEMA, {}) == {}

# Helper functions
def _set_up_flask_logging(app, level=logging.DEBUG):
    cf_logging._SETUP_DONE = False
    flask_logging.init(app, level, custom_fields={'cf1': None, 'cf2': None})


def _user_logging(headers, extra, expected):
    app = Flask(__name__)

    @app.route('/test/user/logging')
    def _logging_correlation_id_route():
        logger, stream = config_logger('user.logging')
        logger.info('in route headers', extra=extra)
        assert check_log_record(stream, JOB_LOG_SCHEMA, expected) == {}
        return Response('ok')

    _set_up_flask_logging(app)
    client = app.test_client()
    _check_expected_response(client.get('/test/user/logging', headers=headers))


def _check_expected_response(response, status_code=200, body='ok'):
    assert response.status_code == status_code
    if body is not None:
        assert response.get_data().decode() == body



""" Module that tests the integration of cf_logging with Sanic """
import logging
import pytest
import sanic
from sanic.response import text
from sap import cf_logging

from sap.cf_logging import sanic_logging
from tests.log_schemas import WEB_LOG_SCHEMA, JOB_LOG_SCHEMA
from tests.common_test_params import v_str, v_num, get_web_record_header_fixtures
from tests.schema_util import extend
from tests.util import (
    check_log_record,
    config_logger,
    enable_sensitive_fields_logging,
)


# pylint: disable=protected-access
@pytest.mark.xfail(raises=TypeError, strict=True)
def test_sanic_requires_valid_app():
    """ Test that the api requires a valid app to be passed in """
    sanic_logging.init({})


FIXTURE = get_web_record_header_fixtures()
FIXTURE.append(({'no-content-length': '1'}, {'response_size_b': v_num(-1)}))


@pytest.fixture(autouse=True)
def before_each():
    """ enable all fields to be logged """
    enable_sensitive_fields_logging()
    yield


@pytest.mark.parametrize("headers, expected", FIXTURE)
def test_sanic_request_log(headers, expected):
    """ Test that the JSON logs contain the expected properties based on the
        input.
    """
    app = sanic.Sanic('test cf_logging')

    @app.route('/test/path')
    async def _headers_route(request):
        if 'no-content-length' in request.headers:
            return text('ok', headers={'Content-Type': 'text/plain'})
        return text('ok', headers={'Content-Length': 2, 'Content-Type': 'text/plain'})

    _set_up_sanic_logging(app)
    _, stream = config_logger('cf.sanic.logger')

    client = app.test_client
    _check_expected_response(client.get(
        '/test/path', headers=headers)[1], 200, 'ok')
    assert check_log_record(stream, WEB_LOG_SCHEMA, expected) == {}


def test_web_log():
    """ Test that custom attributes are logged """
    _user_logging({}, {'myprop': 'myval'}, {'myprop': v_str('myval')}, False)


def test_missing_request():
    """ That the correlation ID when the request is missing """
    _user_logging({'X-CorrelationID': '298ebf9d-be1d-11e7-88ff-2c44fd152860'},
                  {},
                  {'correlation_id': v_str('-')},
                  False)


def test_logs_correlation_id():
    """ Test the setting of the correlation id based on the headers """
    _user_logging({'X-CorrelationID': '298ebf9d-be1d-11e7-88ff-2c44fd152860'},
                  {},
                  {'correlation_id': v_str(
                      '298ebf9d-be1d-11e7-88ff-2c44fd152860')},
                  True)

def test_custom_fields_set():
    """ Test custom fields are set up """
    app = sanic.Sanic('test cf_logging')
    _set_up_sanic_logging(app)
    assert 'cf1' in cf_logging.FRAMEWORK.custom_fields.keys()


# Helper functions
def _set_up_sanic_logging(app, level=logging.DEBUG):
    cf_logging._SETUP_DONE = False
    sanic_logging.init(app, level, custom_fields={'cf1': None})


def _user_logging(headers, extra, expected, provide_request=False):
    app = sanic.Sanic(__name__)

    @app.route('/test/user/logging')
    async def _logging_correlation_id_route(request):
        logger, stream = config_logger('user.logging')
        new_extra = extend(extra, {'request': request}) if provide_request else extra
        logger.info('in route headers', extra=new_extra)
        assert check_log_record(stream, JOB_LOG_SCHEMA, expected) == {}
        return text('ok')

    _set_up_sanic_logging(app)
    client = app.test_client
    _check_expected_response(client.get('/test/user/logging', headers=headers)[1])


def _check_expected_response(response, status_code=200, body=None):
    print(response.text)
    assert response.status == status_code

    if body is not None:
        assert response.text == body



""" Module that tests the integration of cf_logging with Falcon """
import logging
import pytest
import falcon
from falcon import testing
from falcon_auth import FalconAuthMiddleware, BasicAuthBackend
from sap import cf_logging
from sap.cf_logging import falcon_logging
from sap.cf_logging.core.constants import REQUEST_KEY
from tests.log_schemas import WEB_LOG_SCHEMA, JOB_LOG_SCHEMA
from tests.common_test_params import (
    v_str, auth_basic, get_web_record_header_fixtures
)
from tests.util import (
    check_log_record,
    config_logger,
    enable_sensitive_fields_logging
)


# pylint: disable=protected-access, missing-docstring,too-few-public-methods

@pytest.fixture(autouse=True)
def before_each():
    """ enable all fields to be logged """
    enable_sensitive_fields_logging()
    yield


@pytest.mark.xfail(raises=TypeError, strict=True)
def test_falcon_requires_valid_app():
    """ Test the init api expects a valid app """
    falcon_logging.init({})


FIXTURE = get_web_record_header_fixtures()

class TestResource:
    def on_get(self, req, resp): # pylint: disable=unused-argument,no-self-use
        resp.set_header('Content-Type', 'text/plain')
        resp.status = falcon.HTTP_200
        resp.body = 'ok'


@pytest.mark.parametrize('headers, expected', FIXTURE)
def test_falcon_request_log(headers, expected):
    """ That the expected records are logged by the logging library """
    app = falcon.API(middleware=falcon_logging.LoggingMiddleware())
    app.add_route('/test/path', TestResource())

    _set_up_falcon_logging(app)
    _check_falcon_request_log(app, headers, expected)


class User(object):  # pylint: disable=useless-object-inheritance
    def __init__(self, key, name):
        self.key = key
        self.name = name

@pytest.mark.parametrize('user', [
    User(None, 'user'),
    User('custom_username_key', 'new_user')
])
def test_falcon_request_logs_user(user):
    user_dict = dict([(user.key or 'username', user.name)])
    basic_auth = BasicAuthBackend(lambda username, password: user_dict)
    app = falcon.API(middleware=[
        falcon_logging.LoggingMiddleware(),
        FalconAuthMiddleware(basic_auth)
    ])
    app.add_route('/test/path', TestResource())

    args = [app, user.key] if user.key else [app]
    _set_up_falcon_logging(*args)

    expected = {'remote_user': v_str(user.name)}
    _check_falcon_request_log(app, {'Authorization': str(auth_basic(user.name, 'pass'))}, expected)


def _check_falcon_request_log(app, headers, expected):
    _, stream = config_logger('cf.falcon.logger')

    client = testing.TestClient(app)
    _check_expected_response(
        client.simulate_get('/test/path', headers=headers))
    assert check_log_record(stream, WEB_LOG_SCHEMA, expected) == {}


def test_web_log():
    """ That the custom properties are logged """
    _user_logging({}, {'myprop': 'myval'}, {'myprop': v_str('myval')})

def test_logging_without_request():
    """ Test logger is usable in non request context """
    app = falcon.API()
    _set_up_falcon_logging(app)
    cf_logging.FRAMEWORK.context.set_correlation_id('value')

    logger, stream = config_logger('main.logger')
    logger.info('works')
    assert check_log_record(stream, JOB_LOG_SCHEMA, {'msg': v_str('works')}) == {}


def test_correlation_id():
    """ Test the correlation id is logged when coming from the headers """
    _user_logging(
        {'X-Correlation-ID': '298ebf9d-be1d-11e7-88ff-2c44fd152860'},
        {},
        {'correlation_id': v_str('298ebf9d-be1d-11e7-88ff-2c44fd152860')}
    )

def test_custom_fields_set():
    """ Test custom fields are set up """
    app = falcon.API()
    _set_up_falcon_logging(app)
    assert 'cf1' in cf_logging.FRAMEWORK.custom_fields.keys()

# Helper functions
def _set_up_falcon_logging(app, *args):
    cf_logging._SETUP_DONE = False
    falcon_logging.init(app, logging.DEBUG, *args, custom_fields={'cf1': None})


class UserResourceRoute(object):  # pylint: disable=useless-object-inheritance
    def __init__(self, extra, expected):
        self.extra = extra
        self.expected = expected
        self.logger, self.stream = config_logger('user.logging')

    def on_get(self, req, resp):
        self.extra.update({REQUEST_KEY: req})
        self.logger.log(logging.INFO, 'in route headers', extra=self.extra)
        assert check_log_record(self.stream, JOB_LOG_SCHEMA, self.expected) == {}

        resp.set_header('Content-Type', 'text/plain')
        resp.status = falcon.HTTP_200
        resp.body = 'ok'


def _user_logging(headers, extra, expected):
    app = falcon.API(middleware=[
        falcon_logging.LoggingMiddleware()
    ])
    app.add_route('/test/user/logging', UserResourceRoute(extra, expected))
    _set_up_falcon_logging(app)
    client = testing.TestClient(app)
    _check_expected_response(client.simulate_get('/test/user/logging',
                                                 headers=headers))


def _check_expected_response(response, status_code=200, body='ok'):
    assert response.status_code == status_code
    if body is not None:
        assert response.text == body
